import os
import sys
import copy


from tklib.tkobject import tkObject, _analyze_varstr
from tklib.tkutils import pconv, pint, pfloat
from tklib.tkinifile import tkIniFile
import tklib.tkre as tkre

#=========================
# Parameter class
#=========================


class tkParams(tkObject):
    def __init__(self, parameter_file = None, app = None, **args):
#        super(tkObject, self).__init__(**args)
        self._argv = sys.argv
        self._path = parameter_file
        self._app  = app
        self._explanation = {}

        self.update(**args)

    def __del__(self):
#        super(tkObject, self).__del__()
        pass

    def __str__(self):
        return self.ClassPath()


    def dict(self):
        return self.__dict__

    def get_param_dict(self):
        return self.__dict__

    def get_path(self, path):
        if path is not None:
            self._path = path

        return self._path

    def get(self, key, defval = None):
        return self.__dict__.get(key, defval)

    def set(self, key, val, explanation = None):
        setattr(self, key, val)
        self._explanation[key] = explanation
        return val

    def set_attr(self, key, val, explanation = None):
        return self.set(key, val, explanation)

    def get_explanation(self, key):
        expl = self._explanation.get(key, None)
        if expl is None:
            return None

        if self._app:
           return self._app.p(expl)

        return expl
    
    def printinf(self, app):
        print("Parameters:")
        for key in self.__dict__.keys():
            expl = self.get_explanation(key)
            if expl:
                print(f"   {key}: {self.__dict__[key]} ({expl})")
            else:
                print(f"   {key}: {self.__dict__[key]}")

    def print_parameters(self, heading = "", sort_by_keys = True):
        dict = self.get_dict()
        keys = [key for key in dict.keys()]
        if sort_by_keys:
            keys = sorted(keys, key = str.lower)

        print(heading)

        for key in keys:
            if tkre.Match('_', key) or dict[key] is None:
                continue

            expl = self.get_explanation(key)
            if expl:
                print(f"  {key}: {dict[key]}  ({expl})")
            else:
                print(f"  {key}: {dict[key]}")    

    def get_string(self, path = None, section = None, key = None, def_val = None, is_print = False):
        path = self.get_path(path)
        self._path = path

        ini = tkIniFile(IsPrint = IsPrint)
        if ini is None:
            return None

        return ini.get_string(section = section, key = key, def_val = def_val, is_print = is_print)

    def write_string(self, path = None, section = None, key = None, value = None, outfile = None, is_print = False):
        path = self.get_path(path)
        self._path = path

        ini = tkIniFile(IsPrint = IsPrint)
        if ini is None:
            return None

        return write_string(self, section = section, key = key, value = value, is_print = is_print)

    def read_parameters(self, path = None, AddSection = 0, IsPrint = True, follow_vartype = True, read_inifile = False):
        path = self.get_path(path)
        self._path = path

# Store copy of self to check var types
        vars_org = copy.deepcopy(self)

        ini = tkIniFile(IsPrint = IsPrint)
        if ini is None:
            return None

        inf = ini.ReadAll(path, AddSection = AddSection)
        if inf is None:
            return None

# Check list/tuple/dict variables. Get # of elements, and initialize dict variables
        if not read_inifile:
            try:
                del inf['inifile']
            except:
                pass

        keys = inf.keys()
        list_dict = {}
        val_param = None
        for key in keys:
            varname, index = _analyze_varstr(key)
            if index is None:
                continue

            val_param = self.get2(varname, None)
            if type(index) is int:
#                print("*varname=", varname)
                if list_dict.get(index, None) is None:
                    list_dict[varname] = index
                elif list_dict[varname] < index:
                    list_dict[varname] = index
            else:
                self.__dict__[varname] = {}

# Initialize list/tuple variables
        for varname in list_dict.keys():
#            print("**varname=", varname)
            self.__dict__[varname] = [None] * (list_dict[varname] + 1)
#            print(f"{varname}=", self.__dict__[varname])

# Assign variables
        keys = inf.keys()
        for key in keys:
# Check var types from vars_org
            varname, index = _analyze_varstr(key)
#            val_param = vars_org.get2(key, None)
            var = vars_org.__dict__.get(varname, None)
            if var is not None:
# Scalar
                if index is None:
                    val_param = var
# list/tuple/dict
                else:
#                    print("index=", varname, index, type(index))
                    if type(index) is int:
# list/tuple
                        nvar = len(var)
                        if nvar == 0:
                            val_param = ''
                        elif index < nvar:
                            val_param = var[index]
# if vars_org is shorter than the var given
                        else:
                            val_param = var[0]
# dict
                    else:
                        if var.get(index, None) is None:
                            val_param = ""
                        else:
                            val_param = var[index]

            str = inf.get(key, None)
            if str is None:
                continue

            if follow_vartype:
                vtype = type(val_param)
#                print("val=", key, val_param, vtype, str)
                if vtype is float:
                    val = pfloat(str)
                elif vtype is int:
                    val = pint(str)
                elif vtype is bool:
                    if str == 'False' or str == '0' or str == '':
                        val = False
                    else:
                        val = True
                else:
                    val = str
            else:
                val = pconv(str, str)

            self.set_attribute2(key, val)

        return inf

    def save_parameters(self, path = None, section = 'Preferences', 
                update_commandline = True, 
                otherparams = None, other_section = 'OtherParameters',
                sort_by_keys = True, save_inifile = False, IsPrint = True):
        path = self.get_path(path)
        self._path = path

        ini = tkIniFile(path, IsPrint = IsPrint)

        params = self.get_param_dict()
        if update_commandline:
            params['commandline'] = 'python ' + ' '.join(self._argv)

        if not save_inifile:
            try:
                del params['inifile']
            except:
                pass
        keys = [key for key in params.keys()]
        if sort_by_keys:
            keys = sorted(keys, key = str.lower)

        for key in keys:
            if tkre.Match('_', key):
                continue

            val = params.get(key, None)
            ret = True
            if val is None:
                continue
            if type(val) is list or type(val) is tuple:
                for i in range(len(val)):
                    ret = ini.WriteString(section, "{}[{}]".format(key, i), val[i], IsPrint = IsPrint)
                    if not ret:
                        break
            elif type(val) is dict:
                for key2 in val.keys():
                    ret = ini.WriteString(section, "{}[{}]".format(key, key2), val[key2], IsPrint = IsPrint)
                    if not ret:
                        break
            else:
                ret = ini.WriteString(section, key, val, IsPrint = IsPrint)

        if otherparams is not None:
            keys = [key for key in otherparams.keys()]
            if sort_by_keys:
                keys = sorted(keys, key = str.lower)
            for key in keys:
                val = otherparams[key]
                if val is not None:
                    ini.WriteString(other_section, key, val)

        return ret

def main():
    print("")
    print("This is library, not runnable")
    print("")


if __name__ == "__main__":
    main()
